"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Commands_1 = __importDefault(require("../Database/Models/Commands"));
class Command {
    constructor(channels, roles, users, dbRequired = false) {
        this.AllowedChannels = channels;
        this.AllowedRoles = roles;
        this.AllowedUsers = users;
        this.RequiresDatabase = dbRequired;
        this.Modified = false;
    }
    get LocalData() {
        return this.Data[this.Name()];
    }
    set LocalData(value) {
        this.Data[this.Name()] = value;
    }
    get HasLocalData() {
        return !!this.Data[this.Name()];
    }
    HasLocalField(name) {
        return this.HasLocalData && !!this.LocalData[name];
    }
    ValidateChannel(channelId) {
        const AllowedChannels = this.GetValidationPermission('Channel');
        return AllowedChannels.length === 0 || AllowedChannels.includes(channelId);
    }
    ValidateRoles(roles) {
        const AllowedRoles = this.GetValidationPermission('Role');
        return AllowedRoles.length === 0 || roles.find((role) => AllowedRoles.includes(role.id));
    }
    ValidateUsers(user) {
        const AllowedUsers = this.GetValidationPermission('User');
        return AllowedUsers.length === 0 || AllowedUsers.includes(user);
    }
    GetValidationPermission(type) {
        const key = `Allowed${type}s`;
        if (this.HasLocalData) {
            return [...this[key], ...(this.LocalData[key] || [])];
        }
        else {
            return this[key];
        }
    }
    ModifyPermissions(type, action, key, local = false) {
        const permissionKey = `Allowed${type}s`;
        if (!local) {
            if (action === 'add') {
                this[permissionKey].push(key);
            }
            else {
                this[permissionKey] = this[permissionKey].filter(entry => key === entry);
            }
        }
        else if (this.HasLocalField(permissionKey)) {
            if (action === 'add') {
                this.LocalData[permissionKey].push(key);
            }
            else {
                this.LocalData[permissionKey] = this.LocalData[permissionKey].filter(entry => key === entry);
            }
        }
        else {
            this.LocalData[permissionKey] = [key];
        }
        this.Modified = true;
    }
    AddAllowedChannel(channelId, local = false) {
        return this.ModifyPermissions('Channel', 'add', channelId, local);
    }
    AddAllowedRole(roleId, local = false) {
        return this.ModifyPermissions('Role', 'add', roleId, local);
    }
    AddAllowedUser(user, local = false) {
        return this.ModifyPermissions('User', 'add', user, local);
    }
    Call(message) {
        if (!this.ValidateRoles(message.member.roles) || !this.ValidateChannel(message.channel.id)) {
            console.warn('[Failed Permission]', message.member.displayName, message.channel.id, message.content);
            return Promise.resolve('');
        }
        return this
            .Run(message)
            .then(() => this.Save());
    }
    GetContext(message) {
        const parts = message.content.split(' ').slice(1);
        const result = {
            args: []
        };
        parts.forEach((part) => {
            if (part.indexOf('--') === 0) {
                const param = part.split('=');
                result[param[0].replace('--', '')] = param[1] || true;
            }
            else {
                result.args.push(part);
            }
        });
        if (result.args.length < 2) {
            result.args.push('');
            result.args.push('');
        }
        return result;
    }
    RemoveAllowedChannel(channelId, local = false) {
        return this.ModifyPermissions('Channel', 'remove', channelId, local);
    }
    RemoveAllowedRole(roleId, local = false) {
        return this.ModifyPermissions('Role', 'remove', roleId, local);
    }
    RemoveAllowedUsers(user, local = false) {
        return this.ModifyPermissions('User', 'remove', user, local);
    }
    async Save() {
        if (this.RequiresDatabase === false) {
            return Promise.resolve();
        }
        let command = await Commands_1.default.findOne({ Namespace: this.Namespace() });
        if (!command) {
            command = new Commands_1.default({
                Namespace: this.Namespace(),
                AllowedChannels: this.AllowedChannels,
                AllowedRoles: this.AllowedRoles,
                AllowedUsers: this.AllowedUsers,
                Data: {}
            });
        }
        command.Data = Object.assign({}, (command.Data || {}), this.Data);
        if (this.Modified) {
            if (!command.Data[this.Name()]) {
                command.Data[this.Name()] = {};
            }
            command.Data[this.Name()] = {
                AllowedChannels: this.AllowedChannels,
                AllowedRoles: this.AllowedRoles,
                AllowedUsers: this.AllowedUsers
            };
        }
        return command.save();
    }
}
exports.default = Command;
